package dev.kske.chess.board;

import java.util.HashSet;
import java.util.Set;

import dev.kske.chess.board.Piece.Color;

/**
 * Project: <strong>Chess</strong><br>
 * File: <strong>Log.java</strong><br>
 * Created: <strong>09.07.2019</strong><br>
 * Author: <strong>Kai S. K. Engelbart</strong>
 */
public class Log implements Cloneable {

	private MoveNode root, current;

	private Position	enPassant;
	private Color		activeColor;
	private int			fullmoveCounter, halfmoveClock;

	public Log() {
		reset();
	}

	// TODO: Adjust to variations
	@Override
	public Object clone() {
		Log log = null;
		try {
			log = (Log) super.clone();
			if (!isEmpty()) {
				log.current = (MoveNode) current.clone();
				if (current == root) log.root = log.current;
				else log.root = (MoveNode) root.clone();
			}
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return log;
	}

	/**
	 * Adds a move to the move history and adjusts the log to the new position.
	 * 
	 * @param move          The move to log
	 * @param capturedPiece The piece captured with the move
	 * @param pawnMove      {@code true} if the move was made by a pawn
	 */
	public void add(Move move, Piece capturedPiece, boolean pawnMove) {
		enPassant = pawnMove && move.yDist == 2 ? new Position(move.pos.x, move.pos.y + move.ySign) : null;
		if (activeColor == Color.BLACK) ++fullmoveCounter;
		if (pawnMove || capturedPiece != null) halfmoveClock = 0;
		else++halfmoveClock;
		activeColor = activeColor.opposite();
		final MoveNode leaf = new MoveNode(move, capturedPiece, enPassant, activeColor, fullmoveCounter, halfmoveClock);

		if (isEmpty()) {
			root	= leaf;
			current	= leaf;
		} else {
			current.addVariation(leaf);
			current = leaf;
		}
	}

	/**
	 * Removed the last move from the log and adjusts its state to the previous
	 * move.
	 */
	public void removeLast() {
		if (!isEmpty()) {
			if (current.parent == null) {
				current	= null;
				root	= null;
			} else {
				current = current.parent;
				if (!isEmpty()) {
					activeColor		= current.activeColor;
					enPassant		= current.enPassant;
					fullmoveCounter	= current.fullmoveCounter;
					halfmoveClock	= current.halfmoveClock;
				} else reset();
			}
		} else reset();
	}

	public boolean isEmpty() { return root == null; }

	/**
	 * Reverts the log to its initial state corresponding to the default board
	 * position.
	 */
	public void reset() {
		root			= null;
		current			= null;
		enPassant		= null;
		activeColor		= Color.WHITE;
		fullmoveCounter	= 1;
		halfmoveClock	= 0;
	}

	/**
	 * @return The first logged move, or {@code null} if there is none
	 */
	public MoveNode getRoot() { return root; }

	/**
	 * @return the last logged move, or {@code null} if there is none
	 */
	public MoveNode getLast() { return current; }

	public Position getEnPassant() { return enPassant; }

	public void setEnPassant(Position enPassant) { this.enPassant = enPassant; }

	public Color getActiveColor() { return activeColor; }

	public void setActiveColor(Color activeColor) { this.activeColor = activeColor; }

	public int getFullmoveCounter() { return fullmoveCounter; }

	public void setFullmoveCounter(int fullmoveCounter) { this.fullmoveCounter = fullmoveCounter; }

	public int getHalfmoveClock() { return halfmoveClock; }

	public void setHalfmoveClock(int halfmoveClock) { this.halfmoveClock = halfmoveClock; }

	public static class MoveNode {

		public final Move		move;
		public final Piece		capturedPiece;
		public final Position	enPassant;
		public final Color		activeColor;
		public final int		fullmoveCounter, halfmoveClock;

		private MoveNode		parent;
		private Set<MoveNode>	variations	= new HashSet<>();

		public MoveNode(Move move, Piece capturedPiece, Position enPassant, Color activeColor, int fullmoveCounter,
				int halfmoveClock) {
			this.move				= move;
			this.capturedPiece		= capturedPiece;
			this.enPassant			= enPassant;
			this.activeColor		= activeColor;
			this.fullmoveCounter	= fullmoveCounter;
			this.halfmoveClock		= halfmoveClock;
		}

		@Override
		protected Object clone() throws CloneNotSupportedException {
			return super.clone();
		}

		public void addVariation(MoveNode variation) {
			if (!variations.contains(variation)) {
				variations.add(variation);
				variation.parent = this;
			}
		}
	}
}
